--[[
		$Id: AceWardrobe.lua 8140 2006-08-16 05:48:56Z badhex $
		AceWardrobe - An Equipment Management AddOn for World of Warcraft

		Revision Author: 
			Kazun of Warsong

		Requirements:
			Ace

		See AceWardrobe_Readme.txt for more information
]]--

---------------------------------------------------------------------
--  DON'T TOUCH THESE STATIC VALUES UNLESS YOU KNOW WHAT YOU'RE DOING
WARDROBE.VERSION = "3.2.1"
WARDROBE.DBVERSION = 5
WARDROBE.EquipDelay = 0.5		-- This is the speed in which the swap takes place
WARDROBE.SearchDirection = 1		-- This is the direction we are searching bags (needs to be global)
WARDROBE.DblChk = FALSE			-- should we dbl check our inventory
WARDROBE.Count = 0			-- are we busy?
WARDROBE.InCombat = FALSE		-- are we in combat?
WARDROBE.Mounted = FALSE		-- are we mounted?
WARDROBE.LastOutfit = -1		-- last worn outfit before we mounted
WARDROBE.CurrentForm = ""		-- what form we're in
WARDROBE.WearQueue = -1			-- outfit to wear when combat breaks
WARDROBE.InventorySlots = {"AmmoSlot", "HeadSlot","NeckSlot","ShoulderSlot","BackSlot","ChestSlot","ShirtSlot","TabardSlot","WristSlot","HandsSlot","WaistSlot","LegsSlot","FeetSlot","Finger0Slot","Finger1Slot","Trinket0Slot","Trinket1Slot","MainHandSlot","SecondaryHandSlot","RangedSlot"}
WARDROBE.problem_mounts = {
	["Interface\\Icons\\Ability_Mount_PinkTiger"] = 1,
	["Interface\\Icons\\Ability_Mount_WhiteTiger"] = 1,
	["Interface\\Icons\\Spell_Nature_Swiftness"] = 1,
	["Interface\\Icons\\INV_Misc_Foot_Kodo"] = 1,
	["Interface\\Icons\\Ability_Mount_JungleTiger"] =1,
}
---------------------------------------------------------------------

local DEFAULT_OPTIONS = {Version		= 0,
			 Enabled		= TRUE,		-- Default to active
			 Silent			= FALSE,	-- Don't print any messages
			 Current		= -1,		-- This is the current outfit number
			 Mounted		= -1,		-- This is the outfit to use when mounting
			 Default		= -1,		-- This is your default outfit
			 Forms			= { },		-- This table holds each form's outfit number
			 Outfit 		= { }		-- Our outfits
}

AceWardrobe = AceAddon:new({
	name		= WARDROBE.NAME,
	description	= WARDROBE.DESCRIPTION,
	version		= WARDROBE.VERSION,
	aceCompatible 	= "102",
	author		= "Kazun of Warsong",
	category	= ACE_CATEGORY_INVENTORY,
	db		= AceDatabase:new("AceWardrobeDB"),
	defaults	= DEFAULT_OPTIONS,
	cmd		= AceChatCmd:new(WARDROBE.COMMANDS, WARDROBE.CMD_OPTIONS)
})

--------------------------------------------------------------------------------
--  ACE INITILIZATION FUNCTIONS
--  	These functions are required to start and stop the addon
--------------------------------------------------------------------------------

function AceWardrobe:Initialize()
	self.GetOpt = function(loc,var) 
			table.insert(loc, 1, self.profilePath)
			return self.db:get(loc,var)
		      end
	self.SetOpt = function(loc,var,val)
			table.insert(loc, 1, self.profilePath)
			self.db:set(loc,var,val)
		      end
	self.TogOpt = function(loc,var)
			table.insert(loc, 1, self.profilePath)
			return self.db:toggle(self.profilePath,var)
		      end
	self.Debug = FALSE
	self:EventProfileLoaded()
end

-- this function is called when the mod is enabled
function AceWardrobe:Enable()
	self:RegisterZoneEvents()
	self:RegisterCosmos()
end

-- this function is called when the mod is disabled
function AceWardrobe:Disable()
	self:UnregisterZoneEvents()
end

-- these functions contain the register and unregister event code
function AceWardrobe:UnregisterZoneEvents()
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("PLAYER_AURAS_CHANGED")
	self:UnregisterEvent("UPDATE_SHAPESHIFT_FORMS")
	WARDROBE.InCombat = FALSE
end

-- these functions contain the register and unregister event code
function AceWardrobe:RegisterZoneEvents()
	local _, englishClass = UnitClass("player");
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "PlayerEnterCombat")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "PlayerLeaveCombat")
	self:RegisterEvent("PLAYER_AURAS_CHANGED", "PlayerAurasChanged")
	self:RegisterEvent("UPDATE_SHAPESHIFT_FORMS", "PlayerAurasChanged")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "PlayerLeaveCombat")
	self:RegisterEvent("PLAYER_LEAVING_WORLD", "PlayerLeaveCombat")
	self:RegisterEvent("ACE_PROFILE_LOADED", "EventProfileLoaded")
end

-- this event is called when we enter combat
function AceWardrobe:PlayerEnterCombat()
	WARDROBE.InCombat = TRUE
end

-- this event is called when we enter combat
function AceWardrobe:PlayerLeaveCombat()
	WARDROBE.InCombat = FALSE
	if (WARDROBE.WearQueue ~= -1) then
		self:WearOutfit(WARDROBE.WearQueue)
		WARDROBE.WearQueue = -1
	end
end

-- this event is called when an aura changes (used for mounted and stances
function AceWardrobe:PlayerAurasChanged()
	local tmpForm = self:GetForm()
	local _, englishClass = UnitClass("player")
	-----------------------------------------------------------------------
	-- Stance/Form Stuff here
	-----------------------------------------------------------------------
	if (tmpForm == nil) then tmpForm = "" end
	if (WARDROBE.CurrentForm ~= tmpForm) then
		if (englishClass == "ROGUE" and tmpForm == "") then
			WARDROBE.CurrentForm = "No Stealth"
		elseif (englishClass == "DRUID" and tmpForm == "") then
			WARDROBE.CurrentForm = "Normal Form"
		else
			WARDROBE.CurrentForm = tmpForm
		end

		self:DebugPrint("Current Form: ["..WARDROBE.CurrentForm.."]")

		-- lets look for the outfit number for this stance
		local tmpOutfitNum = self.GetOpt({"Forms"}, WARDROBE.CurrentForm)
		if (tmpOutfitNum ~= nil and tmpOutfitNum ~= "" and not self:IsPlayerMounted()) then
			if (englishClass == "DRUID") then
				self:WearOutfit(tmpOutfitNum, TRUE)
			else
				self:WearOutfit(tmpOutfitNum)
			end
		end
	end
	
	-----------------------------------------------------------------------
	-- Mounted Stuff here
	-----------------------------------------------------------------------
	if (self:IsPlayerMounted() == true and WARDROBE.Mounted == FALSE) then
		local mntOutfit = self.GetOpt({ }, "Mounted")
		self:DebugPrint("You have mounted.")
		WARDROBE.Mounted = TRUE
		-- equip your mounted gear
		if (self.GetOpt({ }, "Mounted") ~= -1 and self.GetOpt({ }, "Mounted") ~= nil) then
			if (self.GetOpt({ }, "LastUnmount")) then
				WARDROBE.LastOutfit = self:CurrentOutfit()
			end
			self:WearOutfit(mntOutfit)

		end
	elseif (WARDROBE.Mounted == TRUE and self:IsPlayerMounted() == false) then
		self:DebugPrint("You have dismounted.")
		WARDROBE.Mounted = FALSE
		if (WARDROBE.LastOutfit ~= -1 and self.GetOpt({ }, "LastUnmount")) then
			self:WearOutfit(WARDROBE.LastOutfit)
			WARDROBE.LastOutfit = -1
		else
			if (self.GetOpt({ }, "Default") ~= -1) then
				self:WearOutfit(self.GetOpt({ }, "Default"))
			else
				-- hide me if in silent mode
				if (self.GetOpt({ }, "Silent") == FALSE) then
					self:Print(WARDROBE.ERROR_NODEFAULTOUTFIT)
				end
			end
		end
	end
end

-- this event is called when a new profile is loaded
function AceWardrobe:EventProfileLoaded()
	local tmpVer = self.GetOpt({ }, "Version")

	if (tmpVer == nil or tmpVer < 5) then
		if(tmpVer < 2 or tmpVer == nil) then
			if (tmpVer == nil) then
				self.db:set({self.profilePath[1]}, self.profilePath[2])
				self.db:set(self.profilePath, "Enabled", TRUE)
				self.db:set(self.profilePath, "Current", -1)
				self.db:set(self.profilePath, "Mounted", -1)
				self.db:set(self.profilePath, "Silent", FALSE)
				self.db:set(self.profilePath, "Default", -1)
				self.db:set(self.profilePath, "Forms", { })
				self.db:set(self.profilePath, "Outfit", { })
			end
		end
		if (tmpVer == nil or tmpVer < 3) then
			-- go though all current outfits and set the cloak and helm to the current settings
			for i = 1, table.getn(self.db:get(self.profilePath, "Outfit")) do
				-- set the visability to whatever it is now
				self.SetOpt({"Outfit", i}, "ShowCloak", ShowingCloak())
				self.SetOpt({"Outfit", i}, "ShowHelm", ShowingHelm())
			end
		end
		if (tmpVer == nil or tmpVer < 4) then
			-- go through all current outfits and update the items with links
			for i = 1, table.getn(self.db:get(self.profilePath, "Outfit")) do
				for j = 1, 20 do
					tmpUpName = self.GetOpt({"Outfit", i, "Items"}, j)
					if (tmpUpName ~= "") then
						self.SetOpt({"Outfit", i, "Items"}, j, self:GetBestMatch(tmpUpName))
					end
				end
			end

		end
		if (tmpVer == nil or tmpVer < 5) then
			-- go through all the current outfits and add the unused field
			for i = 1, table.getn(self.db:get(self.profilePath, "Outfit")) do
				for j = 1, 20 do
					self.SetOpt({"Outfit", i, "Unused"}, j, 0)
				end
			end
		end
		self.db:set(self.profilePath, "Version", WARDROBE.DBVERSION)
		self:Print(WARDROBE.INVALID_DB_FORMAT)
		if (tmpVer == nil or tmpVer == 0) then
			StaticPopupDialogs["ACEWARDROBE_FIRSTRUN"] = {
			  text = WARDROBE.FIRST_RUN,
			  button1 = WARDROBE.BUTTON_OK,
			  OnAccept = function()
			      ReloadUI();
			  end,
			  timeout = 0,
			  whileDead = 1,
			  hideOnEscape = 0
			}
			StaticPopup_Show ("ACEWARDROBE_FIRSTRUN")
		end
	end

end

-- this is a placeholder event handle
function AceWardrobe:EventPlaceholder()
	self:DebugPrint("Placeholder Event Fired.", "WARNING")
end

--------------------------------------------------------------------------------
--  COMMAND FUNCTIONS
--  	These functions are fired by the user issuing commands
--------------------------------------------------------------------------------

function strsplit(text)
	local list = {}
	local pos = 1
	local first, last = strfind(text, " ", pos)
	if first then
		tinsert(list, strsub(text, pos, first-1))
		tinsert(list, strsub(text, last+1))
	end
	return list
end

function AceWardrobe:SetOutfitIgnoreSlot(text)
	local args = strsplit(text)
	if (getn(args) < 2) then
		self:Print("Not enough args")
		return
	else
		if(tonumber(args[1]) ~= nil and tonumber(args[1]) > 0 and tonumber(args[1]) < 21) then
			args[1] = tonumber(args[1])
		else
			if (not pcall(GetInventorySlotInfo, args[1])) then
				self:Print(string.format("Unknown slot type [%s]", args[1]))
				return
			else
				args[1] = GetInventorySlotInfo(args[1])+1
			end
		end
	end
	if (1 == 1) then
		local OutfitNumber = self:GetOutfitNumber(args[2])
		if (OutfitNumber == -1) then
			self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
			return
		else
			local opt = self.GetOpt({"Outfit", OutfitNumber, "Unused"}, tonumber(args[1]))
			if (opt == 1) then
				self.SetOpt({"Outfit", OutfitNumber, "Unused"}, tonumber(args[1]), 0)
				self:Print(string.format(WARDROBE.OUTFIT_USLOT, WARDROBE.InventorySlots[args[1]], args[2]))
			else
				self.SetOpt({"Outfit", OutfitNumber, "Unused"}, tonumber(args[1]), 1)
				self:Print(string.format(WARDROBE.OUTFIT_ISLOT, WARDROBE.InventorySlots[args[1]], args[2]))
			end
		end
	end
end

function AceWardrobe:ReturnCheckedStatus(value, i)
	local opt = self.GetOpt({"Outfit", value, "Unused"}, i)
	return opt
end

function AceWardrobe:UpdateOutfit(name)
	local OutfitNumber = self:GetOutfitNumber(name)

	if (OutfitNumber == -1) then
		self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
		return
	else
		-- update the outfit
		local CurrentInventory = self:GetCurrentInventory()
		-- step through the item table and save it in order
		for i = 1, table.getn(CurrentInventory) do
			--self.SetOpt({"Outfit", OutfitNumber, "Items"}, i, CurrentInventory[i])
			tmpName = GetInventoryItemLink("player", GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
			if (tmpName ~= nil) then
				self.SetOpt({"Outfit", OutfitNumber, "Items"}, i, tmpName)
			else
				local hasItem = AceWardrobeTooltip:SetInventoryItem("player", GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
				if (hasItem) then
					ttext = getglobal("AceWardrobeTooltipTextLeft1")
					if (ttext and ttext:IsShown() and ttext:GetText() ~= nil) then
					self.SetOpt({"Outfit", OutfitNumber, "Items"}, i, ttext:GetText())
				end
				else
					self.SetOpt({"Outfit", OutfitNumber, "Items"}, i, "")
				end
			end
		end
		-- update the status of items visability
		self.SetOpt({"Outfit", OutfitNumber}, "ShowCloak", ShowingCloak())
		self.SetOpt({"Outfit", OutfitNumber}, "ShowHelm", ShowingHelm())
		-- now we need to set the current outfit to this saved one
		self.SetOpt({ }, "Current", OutfitNumber)
		self:Print(string.format(WARDROBE.OUTFIT_UPDATED, OutfitNumber))
	end
end

function AceWardrobe:TogNoAutoDefault()
	local NoAuto = self.GetOpt({ }, "NoAuto")

	if (NoAuto == FALSE or NoAuto == nil) then
		self.db:set(self.profilePath, "NoAuto", TRUE)
		self.cmd:msg(WARDROBE.NOAUTO_ENABLE)
	else
		self.db:set(self.profilePath, "NoAuto", FALSE)
		self:Print(WARDROBE.NOAUTO_DISABLE)
	end	
end
function AceWardrobe:ToggleSilentMode()
	local tmpSilent = self.GetOpt({ }, "Silent")

	if (tmpSilent == FALSE or tmpSilent == nil) then
		self.db:set(self.profilePath, "Silent", TRUE)
		self.cmd:msg(WARDROBE.SILENT_ENABLE)
	else
		self.db:set(self.profilePath, "Silent", FALSE)
		self:Print(WARDROBE.SILENT_DISABLE)
	end
end

function AceWardrobe:ToggleLastUnmount()
	local tmpSilent = self.GetOpt({ }, "LastUnmount")

	if (tmpSilent == FALSE or tmpSilent == nil) then
		self.db:set(self.profilePath, "LastUnmount", TRUE)
		self.cmd:msg(WARDROBE.LASTUMOUNT_ENABLE)
	else
		self.db:set(self.profilePath, "LastUnmount", FALSE)
		self:Print(WARDROBE.LASTUMOUNT_DISABLE)
	end
end

function AceWardrobe:ClearForms()
	self.SetOpt({ }, "Forms")
	self.SetOpt({ }, "Forms", { })
	self:Print(WARDROBE.OUTFIT_FORM_CLEARALL)
end

function AceWardrobe:SetOutfitForForm(name)
	local OutfitNumber = self:GetOutfitNumber(name)
	local _, englishClass = UnitClass("player")

	if (WARDROBE.CurrentForm == "") then
		if (englishClass == "ROGUE") then
			WARDROBE.CurrentForm = "No Stealth"
		elseif (englishClass == "DRUID") then
			WARDROBE.CurrentForm = "Normal Form"
		else
			self:Print(WARDROBE.ERROR_NOTINFORM)
			return
		end
	end

	if (name == nil or name == "") then
		self.SetOpt({"Forms"}, WARDROBE.CurrentForm)
		self:Print(WARDROBE.OUTFIT_FORM_CLEAR)
		return
	end	
	if (OutfitNumber == -1) then
		self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
		return
	else
		-- set the outfit number for this form
		self.SetOpt({"Forms"}, WARDROBE.CurrentForm, OutfitNumber)
		self:Print(string.format(WARDROBE.OUTFIT_FORM_SET, OutfitNumber, WARDROBE.CurrentForm))
	end
end

function AceWardrobe:ListAllForms()
	local i,name,form,outfit
	local _, englishClass = UnitClass("player")

	self:Print(WARDROBE.FORM_HEADER)
	if (englishClass == "ROGUE") then
		outfit = self.GetOpt({"Forms"}, "No Stealth")
		if (outfit ~= nil) then
			self:Print(string.format(WARDROBE.FORM_LIST_WITH_OUTFIT, 0, "No Stealth", outfit))
		else
			self:Print(string.format(WARDROBE.FORM_LIST_NO_OUTFIT, 0, "No Stealth"))
		end		
	end
	if (englishClass == "DRUID") then
		outfit = self.GetOpt({"Forms"}, "Normal Form")
		if (outfit ~= nil) then
			self:Print(string.format(WARDROBE.FORM_LIST_WITH_OUTFIT, 0, "Normal Form", outfit))
		else
			self:Print(string.format(WARDROBE.FORM_LIST_NO_OUTFIT, 0, "Normal Form"))
		end		
	end
	for i=1,GetNumShapeshiftForms() do
		_,name,is_active = GetShapeshiftFormInfo(i)
		if (name ~= nil) then
			outfit = self.GetOpt({"Forms"}, name)
			if (outfit ~= nil) then
				self:Print(string.format(WARDROBE.FORM_LIST_WITH_OUTFIT, i, name, outfit))
			else
				self:Print(string.format(WARDROBE.FORM_LIST_NO_OUTFIT, i, name))
			end
		end
	end
end

-- Sets the outfit number to be used when mounted
function AceWardrobe:SetMountedOutfit(inv)
	local tmpDB = { }
	if (inv == nil or inv == "") then
		self.db:set(self.profilePath, "Mounted", -1)
		self:Print(WARDROBE.OUTFIT_MOUNT_CLEAR)
	else
		if (self:GetOutfitNumber(inv) ~= -1) then
			self.db:set(self.profilePath, "Mounted", self:GetOutfitNumber(inv))
		else
			self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
		end
		self:Print(string.format(WARDROBE.OUTFIT_MOUNT_SET, self:GetOutfitNumber(inv)))
	end
end

function AceWardrobe:SetDefaultOutfit(inv)
	local tmpDB = { }
	if (inv == nil or inv == "") then
		self.db:set(self.profilePath, "Default", -1)
		self:Print(WARDROBE.OUTFIT_DEFAULT_CLEAR)
	else
		if (self:GetOutfitNumber(inv)) then
			self.db:set(self.profilePath, "Default", self:GetOutfitNumber(inv))
		else
			self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
		end
		self:Print(string.format(WARDROBE.OUTFIT_DEFAULT_SET, self:GetOutfitNumber(inv)))
	end
end

-- list all the outfits in number order
function AceWardrobe:ListOutfits()
	local OutfitList = self:OutfitList()
	local OutfitCount = table.getn(OutfitList)
	local CurrentOutfit = self:CurrentOutfit()
	local tmpPrinted = FALSE

	self:Print(WARDROBE.LIST_HEADER)

	if (OutfitCount > 0) then
		for i = 1, OutfitCount do
			if(OutfitList[i] ~= "") then
				if (i == CurrentOutfit) then
					self:Print("  >["..i.."] "..OutfitList[i])
				else
					self:Print("    ["..i.."] "..OutfitList[i])
				end
				tmpPrinted = TRUE
			end
		end
	else
		-- we have no outfits
		self:DebugPrint("Current Outfit #: "..CurrentOutfit)
		self:Print("    o "..WARDROBE.ERROR_OUTFIT_NOFIND2)
		return
	end
	if (tmpPrinted == FALSE) then
		self:Print("    o "..WARDROBE.ERROR_OUTFIT_NOFIND2)
		self:DebugPrint("Current Outfit #: "..CurrentOutfit)
	end
end

-- wear an outfit by number or name
function AceWardrobe:WearOutfit(name, force)
	if (WARDROBE.DblChk == FALSE and WARDROBE.Count == 0) then
		if (WARDROBE.InCombat == FALSE or force == TRUE) then
			if (not UnitIsGhost("player")) then
				local tmpSilent = self.GetOpt({ }, "Silent")
				if (tonumber(name) ~= nil and self.GetOpt({"Outfit", tonumber(name)}, "OutfitName") ~= nil) then
					if (tmpSilent == FALSE) then
						self:Print(string.format(WARDROBE.OUTFIT_SWITCH, self.GetOpt({"Outfit", tonumber(name)}, "OutfitName")))
					end
				else
					if (tmpSilent == FALSE) then
						self:Print(string.format(WARDROBE.OUTFIT_SWITCH, name))
					end
				end 
				self:WearOutfit_Next(name)
			else
				self:Print(WARDROBE.ERROR_DEADORGHOST)
			end
		else
			self:WearOutfit_Weapons(name)
			WARDROBE.WearQueue = self:GetOutfitNumber(name)
			self:Print(string.format(WARDROBE.ERROR_INCOMBAT, WARDROBE.WearQueue))
		end
	else
		self:Print(WARDROBE.ERROR_BUSY)
	end
end

function AceWardrobe:WearOutfit_Weapons(name)
	local freeBagSpacesUsed = { };
	local OutfitNumber = self:GetOutfitNumber(name)

	if (OutfitNumber == -1) then
		self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
		return
	else
		local tmpSavedOutfit = self:GetSavedOutfit(OutfitNumber)
		local tmpCurrentOutfit = self:GetCurrentInventory()

		for i = 18, 19 do
			self:DebugPrint("Compare "..tmpCurrentOutfit[i].." -> "..tmpSavedOutfit[i]);
			if (tmpCurrentOutfit[i] == tmpSavedOutfit[i]) then
				self:DebugPrint("     Matched. Ignoring")
			elseif (tmpSavedOutfit[i] == "") then
				self:DebugPrint("     Blank. Ignoring")
			else
				if (tmpCurrentOutfit[18] == tmpSavedOutfit[19] and tmpCurrentOutfit[19] == tmpSavedOutfit[18]) then
					self:SwapInventoryItem(18, 19)
				else 
					if (self:EquipBagItem(tmpSavedOutfit[i], GetInventorySlotInfo(WARDROBE.InventorySlots[i])) == FALSE) then
						self:Print(string.format(WARDROBE.ERROR_ITEM_NOFIND, tmpSavedOutfit[i]))
					end
				end
			end
		end
	end
end

function AceWardrobe:WearOutfit_Next(name)
	local freeBagSpacesUsed = { };
	local OutfitNumber = self:GetOutfitNumber(name)

	if (OutfitNumber == -1) then
		self:Print(WARDROBE.ERROR_OUTFIT_NOFIND)
		return
	else
		self.SetOpt({ }, "Current", OutfitNumber)
		-- time equip the outfit
		local tmpSavedOutfit = self:GetSavedOutfit(OutfitNumber)
		local tmpCurrentOutfit = self:GetCurrentInventory()
		-- update the status of items visability
		ShowCloak(self.GetOpt({"Outfit", OutfitNumber}, "ShowCloak"))
		ShowHelm(self.GetOpt({"Outfit", OutfitNumber}, "ShowHelm"))

		if (tmpSavedOutfit == nil) then return; end
		for i = 1, table.getn(tmpCurrentOutfit) do
			if (tmpSavedOutfit[i] ~= nil) then
				self:DebugPrint("Compare "..tmpCurrentOutfit[i].." -> "..tmpSavedOutfit[i]);
			end
			if (tmpCurrentOutfit[i] == tmpSavedOutfit[i]) then
				self:DebugPrint("     Matched. Ignoring")
			elseif(tmpSavedOutfit[i] == nil or self.GetOpt({"Outfit", OutfitNumber, "Unused"}, i) == 1) then
				self:DebugPrint("     Ignoring Slot")
			elseif(tmpCurrentOutfit[i] ~= "" and tmpSavedOutfit[i] == "") then
				-- we want to bag this item
				if (i == 18) then
					if (tmpSavedOutfit[18] == "" and tmpSavedOutfit[19] == "") then
						-- careful with weapons
						PickupInventoryItem(GetInventorySlotInfo(WARDROBE.InventorySlots[18]))
						result, freeBagSpacesUsed = self:BagItem(freeBagSpacesUsed)
						if (not result) then
							self:Print(string.format(WARDROBE.ERROR_NOROOM, tmpCurrentOutfit[18]))
						end
						PickupInventoryItem(GetInventorySlotInfo(WARDROBE.InventorySlots[19]))
						result, freeBagSpacesUsed = self:BagItem(freeBagSpacesUsed)
						if (not result) then
							self:Print(string.format(WARDROBE.ERROR_NOROOM, tmpCurrentOutfit[19]))
						end
					elseif(tmpCurrentOutfit[18] ~= tmpSavedOutfit[18] and tmpSavedOutfit[19] == "") then
						--nothing
					else
						PickupInventoryItem(GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
						result, freeBagSpacesUsed = self:BagItem(freeBagSpacesUsed)
						if (not result) then
							self:Print(string.format(WARDROBE.ERROR_NOROOM, tmpCurrentOutfit[i]))
						end
					end
				elseif (i == 19) then
					-- nothing
				else
					PickupInventoryItem(GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
					result, freeBagSpacesUsed = self:BagItem(freeBagSpacesUsed)
					if (not result) then
						self:Print(string.format(WARDROBE.ERROR_NOROOM, tmpCurrentOutfit[i]))
					end
				end
			else
				-- we can grab the item right from our bag and expect it to be there
				if (self:EquipBagItem(tmpSavedOutfit[i], GetInventorySlotInfo(WARDROBE.InventorySlots[i])) == FALSE) then
					-- check if we need to preform a swap
					if (WARDROBE.DblChk == FALSE) then
						Timex:AddNamedSchedule(nil, WARDROBE.EquipDelay, false, 3, AceWardrobe["WearOutfit_Next"], AceWardrobe, name)
						WARDROBE.Count = 0
						WARDROBE.DblChk = TRUE
						if (tmpCurrentOutfit[14] == tmpSavedOutfit[15] and tmpCurrentOutfit[15] == tmpSavedOutfit[14]) then
							self:SwapInventoryItem(14, 15)
						end
						if (tmpCurrentOutfit[16] == tmpSavedOutfit[17] and tmpCurrentOutfit[17] == tmpSavedOutfit[16]) then
							self:SwapInventoryItem(16, 17)
						end
						if (tmpCurrentOutfit[18] == tmpSavedOutfit[19] and tmpCurrentOutfit[19] == tmpSavedOutfit[18]) then
							self:SwapInventoryItem(18, 19)
						end
						return
					else
						if (WARDROBE.Count == 2) then
							self:Print(string.format(WARDROBE.ERROR_ITEM_NOFIND, tmpSavedOutfit[i]))
							-- *FIXME* add a check your bank message if item is in there
						end
					end
				end
			end
		end
		if (WARDROBE.DblChk == TRUE and WARDROBE.Count ~= 2) then
			WARDROBE.Count = WARDROBE.Count + 1
		else
			WARDROBE.DblChk = FALSE
			WARDROBE.Count = 0
		end
	end
end

-- saves your current inventory into an outfit
function AceWardrobe:SaveOutfit(name)
	local OutfitNumber = self:GetOutfitNumber(name)

	-- check to see if we already have an outfit named this
	if (OutfitNumber ~= -1) then
		self:Print(string.format(WARDROBE.ERROR_OUTFIT_DUPLICATE, OutfitNumber))
	else
		local FreeSlot = self:GetNextFreeOutfitSlot()
		local CurrentInventory = self:GetCurrentInventory()
		-- our outfit doesn't exists.. save it
		self:Print(string.format(WARDROBE.OUTFIT_SAVE, name, FreeSlot))
		self.SetOpt({"Outfit", FreeSlot}, "OutfitName", name)
		if ((self.GetOpt({ }, "Default") == -1) and not self.GetOpt({ }, "NoAuto")) then
			self.db:set(self.profilePath, "Default", FreeSlot)
			self:Print(WARDROBE.OUTFIT_DEFAULT_SET2)
		end

		-- step through the item table and save it in order
		for i = 1, table.getn(CurrentInventory) do
			--self.SetOpt({"Outfit", FreeSlot, "Items"}, i, CurrentInventory[i])
			tmpName = GetInventoryItemLink("player", GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
			if (tmpName ~= nil) then
				self.SetOpt({"Outfit", FreeSlot, "Items"}, i, tmpName)
				self.SetOpt({"Outfit", FreeSlot, "Unused"}, i, 0)
			else
				local hasItem = AceWardrobeTooltip:SetInventoryItem("player", GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
				if (hasItem) then
					ttext = getglobal("AceWardrobeTooltipTextLeft1")
					if (ttext and ttext:IsShown() and ttext:GetText() ~= nil) then
					self.SetOpt({"Outfit", FreeSlot, "Items"}, i, ttext:GetText())
					self.SetOpt({"Outfit", FreeSlot, "Unused"}, i, 0)
				end
				else
					self.SetOpt({"Outfit", FreeSlot, "Items"}, i, "")
					self.SetOpt({"Outfit", FreeSlot, "Unused"}, i, 0)
				end
			end
		end
		-- update the status of items visability
		self.SetOpt({"Outfit", FreeSlot}, "ShowCloak", ShowingCloak())
		self.SetOpt({"Outfit", FreeSlot}, "ShowHelm", ShowingHelm())

		-- now we need to set the current outfit to this saved one
		self.SetOpt({ }, "Current", FreeSlot)
	end
end

-- delete a slot number
function AceWardrobe:DeleteOutfit(name)
	local OutfitToDelete = self:GetOutfitNumber(name)
	if (OutfitToDelete == -1) then
		self:Print(string.format(WARDROBE.ERROR_OUTFIT_NOFIND3, name))
	else
		self:Print(string.format(WARDROBE.OUTFIT_DELETE, name, OutfitToDelete))
		self:DeleteOutfitBySlot(OutfitToDelete)
		if (self.GetOpt({ }, "Default") == OutfitToDelete) then
			self.db:set(self.profilePath, "Default", -1)
			self:Print(WARDROBE.OUTFIT_DEFAULT_DELETED)
		end
	end
end

-- delete all outfits
function AceWardrobe:ClearAllOutfits()
	self.SetOpt({ }, "Outfit")
	self.SetOpt({ }, "Outfit", { })
	self.SetOpt({ }, "Forms", { })
	self.SetOpt({ }, "Current", -1)
	self.SetOpt({ }, "Mounted", -1)
	self.SetOpt({ }, "Default", -1)
	self:Print(WARDROBE.OUTFIT_ALLCLEAR)
end

--------------------------------------------------------------------------------
--  WORKING FUNCTIONS
--  	These functions do most of the dirty work in the mod
--	These functions will return specific data related to acewardrobe
--------------------------------------------------------------------------------

-- Takes: Table, number, number
-- Returns: boolean
-- See if this bag / slot is being used or is free
function AceWardrobe:UsedThisSlot(freeBagSpacesUsed, theBag, theSlot)
	for i = 1, table.getn(freeBagSpacesUsed) do
		if (freeBagSpacesUsed[i][1] == theBag and freeBagSpacesUsed[i][2] == theSlot) then
			return true
		end
	end
	return false
end

-- Takes: Table
-- Returns: boolean, Table
-- This will put an item on the character into a free bag slot
function AceWardrobe:BagItem(freeBagSpacesUsed)

    -- for each bag and slot
	for theBag = 0, 4, 1 do
		local numSlot = GetContainerNumSlots(theBag)
		for theSlot = 1, numSlot, 1 do

			-- get info about the item here
			local texture, itemCount, locked = GetContainerItemInfo(theBag, theSlot)

			-- if we found nothing, add us to the list of free bag slots
			if (not texture and not self:UsedThisSlot(freeBagSpacesUsed, theBag, theSlot)) then
				PickupContainerItem(theBag,theSlot)
				table.insert(freeBagSpacesUsed, {theBag, theSlot})
				return true, freeBagSpacesUsed
			end
		end
	end

	AutoEquipCursorItem()

	return false, freeBagSpacesUsed
end

-- Takes: string
-- Returns: string
-- This function returns an itemlink currently on your character to the closest possible match
-- returns blank if none found
function AceWardrobe:GetBestMatch(name)
	-- First go through the inventory
	for i = 1, table.getn(WARDROBE.InventorySlots) do
		tmpName = GetInventoryItemLink("player", GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
		_,_,PrintName=string.find(tmpName or "","^.*%[(.*)%].*$")
		if (name == PrintName) then
			return tmpName
		end
	end

	-- Next go through the bags
	bagcounterA = 0
	bagcounterB = NUM_CONTAINER_FRAMES

	for bag = bagcounterA, bagcounterB, 1 do
		local frame = getglobal("ContainerFrame"..bag)
		local counterA = 1
		local counterB = GetContainerNumSlots(bag)

		for slot = counterA, counterB, WARDROBE.SearchDirection do
			-- get the name of the item in this bag/slot
			bagItemName = self:GetItemNameInBagSlot(bag, slot)
			_,_,PrintName=string.find(bagItemName or "","^.*%[(.*)%].*$")
			if (name == PrintName) then
				return bagItemName
			end
		end
	end

    	return ""
end

-- Returns: string
-- This will return the specific stance or form the player is in
function AceWardrobe:GetForm()

	local i,name,form

	for i=1,GetNumShapeshiftForms() do
		_,name,is_active = GetShapeshiftFormInfo(i)
		if is_active then
			form = name
		end
	end
	return form
end

function AceWardrobe:IsPlayerMounted()

	local i,buff
	local mounted = false

	for i=1,24 do
		buff = UnitBuff("player",i)
		if buff then
			if WARDROBE.problem_mounts[buff] or string.find(buff,"QirajiCrystal_") then
				-- hunter could be in group, could be warlock epic mount etc, check if this is truly a mount
				-- or if v1 is set to true, always check every buff. sigh this is slow but really no way around it without more data from users
				AceWardrobeTooltip:SetUnitBuff("player",i)
				if string.find(AceWardrobeTooltipTextLeft2:GetText() or "", "^Increases speed") then
					mounted = true
					i = 25
				end
			elseif string.find(buff,"Mount_") then
				mounted = true
				i = 25
			end
		else
			i = 25
		end
	end

	return mounted
end

-- Takes: number
-- Returns: string
-- This will return the name of a buff
function AceWardrobe:GetBuffName(buffIndex)
	AceWardrobeTooltip:SetUnitBuff("player", buffIndex)
	local Bname = AceWardrobeTooltipTextLeft1:GetText()
	if Bname then return Bname end
end

-- Takes: string
-- This will attempt to equip a bag item by name
function AceWardrobe:EquipBagItem(name, slotNum)
	if (CursorHasItem()) then
		return false
	end
	if (slotNum == GetInventorySlotInfo("AmmoSlot")) then
		name = self:GetBestMatch(name)
	end

	local bag, slot = self:FindContainerItem(name)
	
	if(bag == nil and slot == nil) then
		return FALSE
	else
		self:DebugPrint("        Item found at ["..bag..", "..slot.."]")
		PickupContainerItem(bag, slot)
		if (slotNum < 11 or slotNum == 15 or slotNum == 19) then
			AutoEquipCursorItem()
		else
			EquipCursorItem(slotNum)
		end
		return TRUE
	end
end

-- Takes: number, number
-- Moves an inventory slot
function AceWardrobe:SwapInventoryItem(oNum, dNum)
	if (CursorHasItem()) then
		return false
	end
	PickupInventoryItem(GetInventorySlotInfo(WARDROBE.InventorySlots[dNum]))
	PickupInventoryItem(GetInventorySlotInfo(WARDROBE.InventorySlots[oNum]))
end

-- Takes: string
-- Returns: slot
-- This will return the inventory slot number an item is in
function AceWardrobe:GetInventoryItem(name)
	local tmpCInv = self:GetCurrentInventory()
	
	for i = 1, table.getn(tmpCInv) do
		if (tmpCInv[i] == name) then
			return i-1
		end
	end
	return nil
end

-- Takes: string
-- Returns: number, number
-- This will return the bag number and slot number of an item in your inventory
function AceWardrobe:FindContainerItem(targetItemName)
	local foundBag = nil
	local foundSlot = nil
	self:DebugPrint("    Looking in inventory for "..targetItemName)
    
	-- for each bag and slot
	if ( WARDROBE.SearchDirection == 1 ) then
		bagcounterA = 0
		bagcounterB = NUM_CONTAINER_FRAMES
	else
		bagcounterB = 0
		bagcounterA = NUM_CONTAINER_FRAMES
	end

	for bag = bagcounterA, bagcounterB, WARDROBE.SearchDirection do
		local frame = getglobal("ContainerFrame"..bag)
		local counterA = 1
		local counterB = 1

		if ( WARDROBE.SearchDirection == 1 ) then
			counterA = 1
			counterB = GetContainerNumSlots(bag)
		else
			counterB = 1
			counterA = GetContainerNumSlots(bag)
		end
		for slot = counterA, counterB, WARDROBE.SearchDirection do
			-- get the name of the item in this bag/slot
			bagItemName = self:GetItemNameInBagSlot(bag, slot)
			-- if it matches what we're looking for
			if (targetItemName == bagItemName) then
				foundBag  = bag
				foundSlot = slot
				break
			end
		end
		if (foundBag and foundSlot) then
			break
		end
	end

	if ( WARDROBE.SearchDirection == 1 ) then
		WARDROBE.SearchDirection = -1
	else
		WARDROBE.SearchDirection = 1
	end
    
	return foundBag, foundSlot
end

-- Takes: number, number
-- Returns: string
-- This will return the name of an item in a specific bag slot
function AceWardrobe:GetItemNameInBagSlot(bagNum, slotNum)
    local itemName = ""
    itemName = GetContainerItemLink(bagNum, slotNum)
    if (itemName ~= nil) then
        return itemName
    end
end

-- Returns: string
-- Returns an item name from the tooltip
function AceWardrobe:GetItemNameFromToolTip()

    local itemName = "";
    ttext = getglobal("AceWardrobeTooltipTextLeft1");
    if (ttext and ttext:IsShown() and ttext:GetText() ~= nil) then
        itemName = ttext:GetText();
    end
    
    return itemName;
end

-- Returns: table
-- This will return a table of the current outfits
function AceWardrobe:OutfitList()
	local tempTBL = { }

	for i = 1, table.getn(self.GetOpt({ }, "Outfit")) do
		if (self.GetOpt({"Outfit", i}, "OutfitName") ~= nil) then
			table.insert(tempTBL, self.GetOpt({"Outfit", i}, "OutfitName"))
		else
			table.insert(tempTBL, "")
		end
	end
	return tempTBL
end

-- Returns: table
-- This will return your character's current inventory in order of slot number
function AceWardrobe:GetCurrentInventory()
	local tempTBL = { }
	local tmpName = ""

	for i = 1, table.getn(WARDROBE.InventorySlots) do
		-- get the id number for the item slot
		local id = GetInventorySlotInfo(WARDROBE.InventorySlots[i])
		if (id == GetInventorySlotInfo("AmmoSlot")) then
			local hasItem = AceWardrobeTooltip:SetInventoryItem("player", GetInventorySlotInfo(WARDROBE.InventorySlots[i]))
			if (hasItem) then
				local ttext = getglobal("AceWardrobeTooltipTextLeft1")
				if (ttext and ttext:GetText() ~= nil) then
					tmpName = ttext:GetText()
				else
					tmpName = ""
				end
			end
		else
			-- point the tooltip at this slot
			tmpName = GetInventoryItemLink("player", id)
		end
		if (tmpName ~= nil) then
			table.insert(tempTBL, tmpName)
		else
			table.insert(tempTBL, "")
		end
	end

	return tempTBL
end

-- Takes: number
-- Returns: table
-- This will return a saved outfit table
function AceWardrobe:GetSavedOutfit(number)
	return self.GetOpt({"Outfit", number}, "Items")
end

-- Returns: number
-- This will return the current outfit number (-1 for unknown)
function AceWardrobe:CurrentOutfit()
	return self.GetOpt({ }, "Current")
end

-- Returns: string
-- This will return the current outfit number (-1 for unknown)
function AceWardrobe:CurrentOutfitName()
	local ofnumber = AceWardrobe.GetOpt({ }, "Current")
	if (ofnumber == -1) then
		return "<No Outfit>"
	else
		return self.GetOpt({"Outfit", ofnumber}, "OutfitName")
	end
end

-- Takes: string
-- Returns: number
-- This will return the outfit number if an outfit is found.  Otherwise returns -1
function AceWardrobe:GetOutfitNumber(name)
	local OutfitList = self:OutfitList()
	local OutfitCount = table.getn(OutfitList)

	for i = 1, OutfitCount do
		if (OutfitList[i] == name) then
			return i
		end
	end
	self:DebugPrint("Checking if number: "..name)
	-- if not by name then lets see if we can find the number
	if (OutfitList[tonumber(name)] ~= "" and OutfitList[tonumber(name)] ~= nil) then
		self:DebugPrint("returning: "..tonumber(name))
		return tonumber(name) 
	end
	return -1
end

-- Returns: number
-- This function will return the next available open slot
function AceWardrobe:GetNextFreeOutfitSlot()
	self:DebugPrint("Looking for free slot")
	local iTop = table.getn(self.GetOpt({ }, "Outfit"))
	for i = 1, iTop do
		if (self.GetOpt({"Outfit", i}, "OutfitName") == nil) then
			return i
		end
	end
	return iTop+1
end

--------------------------------------------------------------------------------
--  UTILITY FUNCTIONS
--  	These functions will be used throughout the mod.
--------------------------------------------------------------------------------

-- delete the supplied slot number
function AceWardrobe:DeleteOutfitBySlot(slotNum)
	self.SetOpt({"Outfit"}, slotNum)
	self.SetOpt({"Outfit"}, slotNum, { })
	if (self.GetOpt({ }, "Current") == slotNum) then self.SetOpt({ }, "Current", -1) end
	if (self.GetOpt({ }, "Mounted") == slotNum) then self.SetOpt({ }, "Mounted", -1) end
	-- We need to go though the forms table and see if we need to reset any
end

-- look up the name of the item in a slot number
function AceWardrobe:GetItemNameAtInventorySlotNumber(slotNum)
    local itemName = ""
    
    --self:DebugPrint(slotNum)
    self:DebugPrint("Slot ["..slotNum.."] is ["..WARDROBE.InventorySlots[slotNum].."].")

    -- get the id number for the item slot
    local id = GetInventorySlotInfo(WARDROBE.InventorySlots[slotNum])

    -- point the tooltip at this slot
    local hasItem = WardrobeTooltip:SetInventoryItem("player", id)
    if (hasItem) then
            ttext = getglobal("WardrobeTooltipTextLeft1")
	    if (ttext and ttext:IsShown() and ttext:GetText() ~= nil) then
	        itemName = ttext:GetText()
	    end
    end
    
    return itemName
end

-- used for normal messages
function AceWardrobe:Print(msg)
	if (msg == nil) then msg = "" end
	self.cmd:msg(msg)
end

-- used for debug messages
function AceWardrobe:DebugPrint(msg, type)
	if(type == nil) then type = "" else type = type..": " end
	if (msg == nil) then msg = "" end

	if (self.Debug) then
		self.cmd:msg(type..msg)
	end
end

--------------------------------------------------------------------------------
--  REGISTRATION FUNCTIONS
--  	These functions will be used to register the mod with cosmos or titan panel.
--------------------------------------------------------------------------------
---------------------------------------------------------------------------------
-- Register with Cosmos UI
---------------------------------------------------------------------------------
function AceWardrobe:RegisterCosmos()

    if (Cosmos_RegisterConfiguration) then 

        Cosmos_RegisterConfiguration(
            "COS_ACEWARDROBE",
            "SECTION",
            WARDROBE.CONFIG_HEADER,
            WARDROBE.CONFIG_HEADER_INFO
        );
        Cosmos_RegisterConfiguration(
            "COS_ACEWARDROBE_HEADER",
            "SEPARATOR",
            WARDROBE.CONFIG_HEADER,
            WARDROBE.CONFIG_HEADER_INFO
        );        Cosmos_RegisterChatCommand (
            "ACEWARDROBE_COMMANDS", -- Some Unique Group ID
            WARDROBE.COMMANDS, -- The Commands
            _,
            "" -- Description String
        );
    end
end

-- Finally register the mod with ace
AceWardrobe:RegisterForLoad()
